import os
import re
import json
import random
import textwrap
import time
import numpy as np
import pickle
import asyncio
import ssl

from PIL import Image, ImageDraw, ImageFont

import praw
import requests
import edge_tts
import whisper

from moviepy.editor import (VideoFileClip, AudioFileClip, TextClip,
                            CompositeVideoClip, concatenate_audioclips, CompositeAudioClip)
from moviepy.video.tools.subtitles import SubtitlesClip
from moviepy.config import change_settings

from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.http import MediaFileUpload

# Zorg dat ImageMagick en FFMPEG goed ingesteld zijn
change_settings({"IMAGEMAGICK_BINARY": r"C:\Program Files\ImageMagick-7.1.1-Q16-HDRI\magick.exe"})
os.environ["FFMPEG_BINARY"] = r"C:\ffmpeg\ffmpeg-master-latest-win64-gpl-shared\bin\ffmpeg.exe"

# ---------------------- CONSTANTEN ----------------------
# De video is al in 9:16; dus geen extra resizing
TIKTOK_RESOLUTION = (1080, 1920)
FORBIDDEN_WORDS = {
    r'\bfuck\b': 'f ck',
    r'\bnaked\b': 'n k d',
    r'\bshit\b': 'sh t',
    r'\bass\b': 'ahh',
    r'\bbitch\b': 'b tch',
    r'\bdick\b': 'd ck',
    r'\basshole\b': 'ahol',
    r'\bfucking\b': 'fricking'
}

REDDIT_CONFIG = {
    "client_id": "hLyJraqTqoqAA-bIp5y9Iw",
    "client_secret": "liVAhHciAxwh8sAaQVV1q-40DhXIlw",
    "user_agent": "script:StoryBot:v1.0 (by /u/Sure-Payment-7884)",
    "subreddit": "AmItheAsshole"
}

VIDEO_CONFIG = {
    "gameplay_path": "gameplay.mp4",
    "output_path": "final_video.mp4",
    "font_size": 80,
    "text_color": "black",
    "font": "Arial-Bold",
    "stroke_color": "white",
    "stroke_width": 2,
    "text_position": ("center", "center")
}

# YouTube upload configuratie
YOUTUBE_CLIENT_SECRET_FILE = "credentials.json"
YOUTUBE_SCOPES = ["https://www.googleapis.com/auth/youtube.upload"]
TOKEN_FILE = "token.pickle"

# (Optioneel) TikTok upload configuratie – pas aan met jouw gegevens
TIKTOK_CLIENT_KEY = "YOUR_TIKTOK_CLIENT_KEY"
TIKTOK_CLIENT_SECRET = "YOUR_TIKTOK_CLIENT_SECRET"

# Globale set voor unieke Reddit-posts
USED_POSTS = set()

# ---------------------- GOOGLE OAUTH TOKEN CACHING ----------------------
def get_authenticated_service():
    """Haalt de YouTube service op met token caching, zodat de OAuth flow niet telkens verschijnt."""
    creds = None
    if os.path.exists(TOKEN_FILE):
        with open(TOKEN_FILE, "rb") as token:
            creds = pickle.load(token)
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            from google.auth.transport.requests import Request
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(YOUTUBE_CLIENT_SECRET_FILE, YOUTUBE_SCOPES)
            creds = flow.run_local_server(port=0)
        with open(TOKEN_FILE, "wb") as token:
            pickle.dump(creds, token)
    return build("youtube", "v3", credentials=creds)

# ---------------------- CORE FUNCTIES ----------------------
def get_reddit_post():
    """Haalt een willekeurige post op met geavanceerde filtering en zorgt voor unieke verhalen."""
    reddit = praw.Reddit(**REDDIT_CONFIG)
    subreddit = reddit.subreddit(REDDIT_CONFIG["subreddit"])
    
    eligible_posts = []
    for post in subreddit.top(time_filter="week", limit=150):
        # Sla posts over die al gebruikt zijn
        if post.id in USED_POSTS:
            continue
        conditions = (
            post.title.lower().startswith("aita for"),
            post.score >= 7000,
            not post.stickied,
            len(post.selftext) > 300,
            "[removed]" not in post.selftext.lower(),
            "[deleted]" not in post.selftext.lower()
        )
        if all(conditions):
            eligible_posts.append(post)
    
    if not eligible_posts:
        # Reset de set als geen nieuwe posts gevonden worden
        USED_POSTS.clear()
        return get_reddit_post()
    
    selected_post = random.choice(eligible_posts)
    USED_POSTS.add(selected_post.id)
    return f"{selected_post.title}\n\n{selected_post.selftext}"

def censor_text(text):
    """Voert tekstcensuur uit op basis van gedefinieerde patronen."""
    for pattern, replacement in FORBIDDEN_WORDS.items():
        text = re.sub(pattern, replacement, text, flags=re.IGNORECASE)
    return text

async def generate_audio(text):
    """
    Genereert spraak met edge-tts en slaat deze op als 'audio.mp3'.
    edge-tts levert natuurlijk klinkende stemmen (in dit voorbeeld "en-US-GuyNeural").
    """
    audio_path = "audio.mp3"
    communicate = edge_tts.Communicate(text, "en-US-GuyNeural")
    await communicate.save(audio_path)
    return audio_path

def transcribe_audio(audio_path):
    """
    Gebruikt Whisper om de audio te transcriberen met woordnauwkeurige timestamps.
    Verwacht dat het model de parameter 'word_timestamps' ondersteunt.
    """
    model = whisper.load_model("base")
    result = model.transcribe(audio_path, word_timestamps=True, task="transcribe")
    
    word_timings = []
    for segment in result["segments"]:
        for word_info in segment.get("words", []):
            word = word_info["word"].strip()
            if word:
                word_timings.append(((word_info["start"], word_info["end"]), word))
    return word_timings

def create_subtitles_clip(word_timings):
    """Genereert een SubtitlesClip op basis van de woord-timings."""
    def text_generator(txt):
        return TextClip(
            txt,
            fontsize=VIDEO_CONFIG["font_size"],
            font=VIDEO_CONFIG["font"],
            color=VIDEO_CONFIG["text_color"],
            stroke_color=VIDEO_CONFIG["stroke_color"],
            stroke_width=VIDEO_CONFIG["stroke_width"],
            size=(int(TIKTOK_RESOLUTION[0] * 0.9), None),
            method="caption",
            align="center"
        )
    return SubtitlesClip(word_timings, text_generator).set_position(VIDEO_CONFIG["text_position"])

def create_video():
    """Verwerkt de video (geen extra resizing, de video is al 9:16)."""
    if os.path.exists(VIDEO_CONFIG["output_path"]):
        os.remove(VIDEO_CONFIG["output_path"])
    
    # Stap 1: Content generatie
    story_text = censor_text(get_reddit_post())
    audio_path = asyncio.run(generate_audio(story_text))
    
    # Stap 2: Media processing
    video = VideoFileClip(VIDEO_CONFIG["gameplay_path"])
    tts_audio = AudioFileClip(audio_path)
    
    # Voeg achtergrondmuziek toe als "muziek.mp3" bestaat
    if os.path.exists("muziek.mp3"):
        bg_music = AudioFileClip("muziek.mp3").volumex(0.2).set_duration(tts_audio.duration)
        combined_audio = CompositeAudioClip([tts_audio, bg_music])
    else:
        combined_audio = tts_audio

    video = video.subclip(0, combined_audio.duration)
    
    # Stap 3: Subtiteling
    word_timings = transcribe_audio(audio_path)
    subtitles = create_subtitles_clip(word_timings).set_duration(combined_audio.duration)
    
    # Stap 4: Compositing
    final = CompositeVideoClip([video, subtitles]).set_audio(combined_audio)
    final.write_videofile(
        VIDEO_CONFIG["output_path"],
        codec="libx264",
        audio_codec="aac",
        fps=24,
        threads=4,
        logger="bar"
    )
    
    # Opruimen
    video.close()
    tts_audio.close()
    if os.path.exists(audio_path):
        os.remove(audio_path)

def upload_to_youtube(video_path, title, description):
    """Uploadt de video naar YouTube via de YouTube API met herbruikbare credentials en een retry-mechanisme."""
    youtube = get_authenticated_service()
    attempts = 3
    for attempt in range(attempts):
        try:
            request = youtube.videos().insert(
                part="snippet,status",
                body={
                    "snippet": {
                        "title": title,
                        "description": description,
                        "categoryId": "28"  # Pas indien nodig aan
                    },
                    "status": {"privacyStatus": "public"}
                },
                media_body=MediaFileUpload(video_path)
            )
            response = request.execute()
            print("YouTube upload response:", response)
            return
        except ssl.SSLWantWriteError as e:
            print(f"SSLWantWriteError op poging {attempt+1}: {e}. Wachten 5 seconden en opnieuw proberen...")
            time.sleep(5)
    raise Exception("Upload naar YouTube is mislukt na meerdere pogingen.")

# (Optioneel) Functie voor TikTok-upload (voorbeeld – pas aan met jouw API-integratie)
def upload_to_tiktok(video_path, caption, client_key, client_secret):
    auth_url = "https://open-api.tiktok.com/oauth/access_token/"  # Voorbeeldendpoint
    auth_payload = {
        "client_key": client_key,
        "client_secret": client_secret,
        "grant_type": "client_credentials"
    }
    auth_response = requests.post(auth_url, data=auth_payload)
    if auth_response.status_code != 200:
        raise Exception("TikTok authenticatie mislukt")
    access_token = auth_response.json().get("data", {}).get("access_token")
    if not access_token:
        raise Exception("Geen TikTok access token ontvangen")
    
    upload_url = "https://open-api.tiktok.com/video/upload/"  # Voorbeeldendpoint
    files = {
        'video': open(video_path, 'rb')
    }
    data = {
        "access_token": access_token,
        "caption": caption,
    }
    response = requests.post(upload_url, data=data, files=files)
    print("TikTok upload response:", response.json())

# ---------------------- MAIN EXECUTIE (CONTINU) ----------------------
if __name__ == "__main__":
    try:
        while True:
            start_time = time.time()
            print("\n--- Nieuwe iteratie gestart ---")
            create_video()
            print(f"✅ Video succesvol gegenereerd in {time.time()-start_time:.2f} seconden")
            
            # Upload naar YouTube
            yt_title = "AITA? #shorts #redditstory #reddit #story"
            yt_description = "AITA? #shorts #redditstory #reddit #story"
            upload_to_youtube(VIDEO_CONFIG["output_path"], yt_title, yt_description)
            
            # (Optioneel) Upload naar TikTok – pas credentials en endpoints aan
            # tiktok_caption = "AITA? #shorts #redditstory #reddit #story"
            # upload_to_tiktok(VIDEO_CONFIG["output_path"], tiktok_caption, TIKTOK_CLIENT_KEY, TIKTOK_CLIENT_SECRET)
            
            # Wacht even voordat een nieuwe iteratie begint (pas de wachttijd aan, bv. 10 minuten = 600 seconden)
            print("Wachten op de volgende iteratie...")
            time.sleep(10)  # Hier ingesteld op 600 seconden (10 minuten)
    except KeyboardInterrupt:
        print("\nStop gevraagd door gebruiker. Programma beëindigen.")
    except Exception as e:
        print(f"❌ Kritieke fout: {str(e)}")
        raise




# tekst op de video
# uplaoden naar youtube, tiktok, insta
# andere tts stem
# formaat video
